/*
 *	$Id: blz2sc.c
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/device.h>
#include <scsi/scsi_all.h>
#include <scsi/scsiconf.h>
#include <vm/vm.h>
#include <vm/vm_kern.h>
#include <vm/vm_page.h>
#include <machine/pmap.h>
#include <amiga/amiga/custom.h>
#include <amiga/amiga/cc.h>
#include <amiga/amiga/device.h>
#include <amiga/amiga/isr.h>
#include <amiga/dev/fasreg.h>
#include <amiga/dev/fasvar.h>
#include <amiga/dev/zthreebusvar.h>
#include <amiga/dev/blz2screg.h>
#include <amiga/dev/blz2scvar.h>

int blz2scprint __P((void *auxp, char *));
void blz2scattach __P((struct device *, struct device *, void *));
int blz2scmatch __P((struct device *, struct cfdata *, void *));

struct scsi_adapter blz2sc_scsiswitch = {
	fas_scsicmd,
	fas_minphys,
	0,			/* no lun support */
	0,			/* no lun support */
	fas_adinfo,
	"blz2sc",
};

struct scsi_device blz2sc_scsidev = {
	NULL,		/* use default error handler */
	NULL,		/* do not have a start functio */
	NULL,		/* have no async handler */
	NULL,		/* Use default done routine */
	"blz2sc",
	0,
};

struct cfdriver blz2sccd = {
	NULL, "blz2sc", (cfmatch_t)blz2scmatch, blz2scattach, 
	DV_DULL, sizeof(struct blz2sc_softc), NULL, 0 };

int blz2sc_intr		__P((struct fas_softc *dev));
int blz2sc_setup_dma	__P((struct fas_softc *sc, void *ptr, int len,
			     int mode));
int blz2sc_build_dma_chain __P((struct fas_softc *sc,
				struct fas_dma_chain *chain, void *p, int l));
int blz2sc_need_bump	__P((struct fas_softc *sc, void *ptr, int len));
void blz2sc_led_dummy	__P((struct fas_softc *sc));

/*
 * if we are an Advanced Systems & Software FastlaneZ3
 */
int blz2scmatch(struct device *pdp, struct cfdata *cdp,	void *auxp)
{
  struct zbus_args *zap;

  if (!is_a1200())
    return(0);

  zap = auxp;
  if (zap->manid == 0x2140 && zap->prodid == 11)
    return(1);

  return(0);
}

void blz2scattach(struct device *pdp, struct device *dp, void *auxp)
{
  struct blz2sc_softc	*sc;
  struct zbus_args	*zap;
  blz2sc_regmap_p	 rp;
  u_int			*pte, page;
  vu_char		*fas;

  zap = auxp;
  fas = (vu_char *)(((char *)zap->va)+0x10000);

  sc = (struct blz2sc_softc *)dp;
  rp = &sc->sc_regmap;

  rp->FAS216.fas_tc_low   = &fas[0x00];
  rp->FAS216.fas_tc_mid   = &fas[0x02];
  rp->FAS216.fas_fifo     = &fas[0x04];
  rp->FAS216.fas_command  = &fas[0x06];
  rp->FAS216.fas_dest_id  = &fas[0x08];
  rp->FAS216.fas_timeout  = &fas[0x0A];
  rp->FAS216.fas_syncper  = &fas[0x0C];
  rp->FAS216.fas_syncoff  = &fas[0x0E];
  rp->FAS216.fas_config1  = &fas[0x10];
  rp->FAS216.fas_clkconv  = &fas[0x12];
  rp->FAS216.fas_test     = &fas[0x14];
  rp->FAS216.fas_config2  = &fas[0x16];
  rp->FAS216.fas_config3  = &fas[0x18];
  rp->FAS216.fas_tc_high  = &fas[0x1C];
  rp->FAS216.fas_fifo_bot = &fas[0x1E];
  rp->cclkaddr		  = &fas[0x21];
  rp->epowaddr		  = &fas[0x31];

  sc->sc_softc.sc_fas  = (fas_regmap_p)rp;
  sc->sc_softc.sc_spec = 0;

  sc->sc_softc.sc_setup_dma	 = blz2sc_setup_dma;
  sc->sc_softc.sc_build_dma_chain= blz2sc_build_dma_chain;
  sc->sc_softc.sc_need_bump	 = blz2sc_need_bump;

  sc->sc_softc.sc_led		 = blz2sc_led_dummy;

  sc->sc_softc.sc_clock_freq   = 40;	/* BlizzardII 1230 runs at 40MHz? */
  sc->sc_softc.sc_timeout      = 250;	/* Set default timeout to 250ms */
  sc->sc_softc.sc_config_flags = 0;
  sc->sc_softc.sc_host_id      = 7;

  sc->sc_softc.sc_bump_sz = NBPG;
  sc->sc_softc.sc_bump_pa = 0x0;

  fasinitialize((struct fas_softc *)sc);

  sc->sc_softc.sc_link.adapter_softc = sc;
  sc->sc_softc.sc_link.adapter_targ = sc->sc_softc.sc_host_id;
  sc->sc_softc.sc_link.adapter = &blz2sc_scsiswitch;
  sc->sc_softc.sc_link.device = &blz2sc_scsidev;

  custom.intreq = INTF_PORTS;
  custom.intena = INTF_SETCLR | INTF_PORTS;

  printf("\n");

/* attach all scsi units on us */
  config_found(dp, &sc->sc_softc.sc_link, blz2scprint);
}

/* print diag if pnp is NULL else just extra */
int blz2scprint(void *auxp, char *pnp)
{
  if (pnp == NULL)
    return(UNCONF);

  return(QUIET);
}

int blz2sc_intr()
{
  blz2sc_regmap_p	 rp;
  struct fas_softc	*dev;
  int			 quickints, i, found;

  found = 0;
  for(i=0; i<blz2sccd.cd_ndevs; i++)
    {
      dev = blz2sccd.cd_devs[i];
      if (!dev)
	continue;

      rp = (blz2sc_regmap_p)dev->sc_fas;

      if (!(*rp->FAS216.fas_status & FAS_STAT_INTERRUPT_PENDING))
	continue;

      found++;

      quickints = 16;
      do
	{
	  dev->sc_status = *rp->FAS216.fas_status;
	  dev->sc_interrupt = *rp->FAS216.fas_interrupt;
      
	  if (dev->sc_interrupt & FAS_INT_RESELECTED)
	    {
	      dev->sc_resel[0] = *rp->FAS216.fas_fifo;
	      dev->sc_resel[1] = *rp->FAS216.fas_fifo;
	    }
	  fasintr(dev);
	} while((*rp->FAS216.fas_status & FAS_STAT_INTERRUPT_PENDING) &&
		--quickints);
    }

  return(found);
}

/* --------- */
void blz2sc_set_dma_adr(struct fas_softc *sc, void *ptr, int mode)
{
  blz2sc_regmap_p	rp;
  unsigned long		p;

  rp = (blz2sc_regmap_p)sc->sc_fas;

  p = ((unsigned long)p)>>1;

  if (mode == FAS_DMA_WRITE)
    p |= BLZ2SC_DMA_WRITE;
  else
    p |= BLZ2SC_DMA_WRITE;

  *rp->epowaddr = (u_char)(p>>24) & 0xFF;
  *rp->cclkaddr = (u_char)(p>>16) & 0xFF;
  *rp->cclkaddr = (u_char)(p>> 8) & 0xFF;
  *rp->cclkaddr = (u_char)(p    ) & 0xFF;
}

/* Set DMA transfer counter */
void blz2sc_set_dma_tc(struct fas_softc *sc, unsigned int len)
{
  *sc->sc_fas->fas_tc_low  = len; len >>= 8;
  *sc->sc_fas->fas_tc_mid  = len; len >>= 8;
  *sc->sc_fas->fas_tc_high = len;
}

/* Initialize DMA for transfer */
int blz2sc_setup_dma(struct fas_softc *sc, void *ptr, int len, int mode)
{
  int	retval;

  retval = 0;

  switch(mode)
    {
    case FAS_DMA_READ:
    case FAS_DMA_WRITE:
      blz2sc_set_dma_adr(sc, ptr, mode);
      blz2sc_set_dma_tc(sc, len);
      break;
    case FAS_DMA_CLEAR:
    default:
      retval = (*sc->sc_fas->fas_tc_high << 16) |
	       (*sc->sc_fas->fas_tc_mid  <<  8) |
		*sc->sc_fas->fas_tc_low;
      
      blz2sc_set_dma_tc(sc, 0);
      break;
    }

  return(retval);
}

/* Check if address and len is ok for DMA transfer */
int blz2sc_need_bump(struct fas_softc *sc, void *ptr, int len)
{
  int	p;

  p = (int)p & 0x03;

  if (p)
    {
      p = 4-p;

      if (len < 256)
	p = len;
    }

  return(p);
}

/* Interrupt driven routines */
int blz2sc_build_dma_chain(struct fas_softc *sc, struct fas_dma_chain *chain,
			   void *p, int l)
{
  int	n;

  if (!l)
    return(0);

#define set_link(n, p, l, f)\
do { chain[n].ptr = (p); chain[n].len = (l); chain[n++].flg = (f); } while(0)

  n = 0;

  if (l < 512)
    {
      set_link(n, (vm_offset_t)p, l, FAS_CHAIN_BUMP);
    }
  else if (
#ifdef M68040
	   (cpu040 && ((vm_offset_t)p >= 0xFFFC0000)) &&
#endif
	   ((vm_offset_t)p >= 0xFF000000))
    {
      int	len;

      while(l)
	{
	  len = ((l > sc->sc_bump_sz) ? sc->sc_bump_sz : l);

	  set_link(n, (vm_offset_t)p, len, FAS_CHAIN_BUMP);

	  p += len;
	  l -= len;
	}
    }
  else
    {
      char		*ptr;
      vm_offset_t	 pa, lastpa;
      int		 len,  prelen,  postlen, max_t;

      ptr = p;
      len = l;

      pa = kvtop(ptr);
      prelen = ((int)ptr & 0x03);

      if (prelen)
	{
	  prelen = 4-prelen;
	  set_link(n, (vm_offset_t)ptr, prelen, FAS_CHAIN_BUMP);
	  ptr += prelen;
	  len -= prelen;
	}

      lastpa = 0;
      while(len > 3)
	{
	  pa = kvtop(ptr);
	  max_t = NBPG - (pa & PGOFSET);
	  if (max_t > len)
	    max_t = len;
	  
	  max_t &= ~3;
	  
	  if (lastpa == pa)
	    sc->sc_chain[n-1].len += max_t;
	  else
	    set_link(n, pa, max_t, FAS_CHAIN_DMA);
	  
	  lastpa = pa+max_t;
	  
	  ptr += max_t;
	  len -= max_t;
	}
      
      if (len)
	set_link(n, (vm_offset_t)ptr, len, FAS_CHAIN_BUMP);
    }

  return(n);
}

/* Turn on led */
void blz2sc_led_dummy(struct fas_softc *sc)
{
}
