#ifndef _FASVAR_H_
#define _FASVAR_H_

/*
 * Do not define if we don't need the vm FIX
 */
#define FAS_NEED_VM_PATCH	1

/*
 * MAXCHAIN is the anticipated maximum number of chain blocks needed. This
 * assumes that we are NEVER requested to transfer more than MAXPHYS bytes.
 */
#define MAXCHAIN	(MAXPHYS/NBPG+2)

/*
 * Maximum number of requests standing by. Could be anything, but I think 9
 * looks nice :-) NOTE: This does NOT include requests already started!
 */
#define MAXPENDING	9	/* 7 IDs + 2 extra */

/*
 * DMA chain block. If flg == FAS_CHAIN_PRG or flg == FAS_CHAIN_BUMP then
 * ptr is a VIRTUAL adress. If flg == FAS_CHAIN_DMA then ptr is a PHYSICAL
 * adress.
 */
struct	fas_dma_chain {
	vm_offset_t	ptr;
	u_short		len;
	short		flg;
};
#define FAS_CHAIN_DMA	0x00
#define FAS_CHAIN_BUMP	0x01
#define FAS_CHAIN_PRG	0x02

/*
 * This is the data that we need for the vm FIX. offset is the offset into
 * the first page. pa is the physical adress of each page. pages is the number
 * of slots used in pa. If pages == 0 the we don't need the FIX right now.
 */
#ifdef FAS_NEED_VM_PATCH
struct	vm_link_data {
	vm_offset_t	pa[MAXCHAIN];
	short		offset;
	short		pages;
};
#endif

/*
 * This struct contains the necessary info for a pending request. Pointer to
 * a scsi_xfer struct and an optional vm_link_data.
 */
struct	fas_pending {
	TAILQ_ENTRY(fas_pending) link;
	struct scsi_xfer	*xs;
#ifdef FAS_NEED_VM_PATCH
	struct vm_link_data	 vm_link_data;
#endif
};

/*
 * nexus contains all active data for one SCSI unit. Parts of the info in this
 * struct survives between scsi commands.
 */
struct nexus {
	struct	scsi_xfer 	*xs;		/* Pointer to request */

	u_char			 ID;		/* ID message to be sent */
	u_char			 clen;		/* scsi command length + */
	u_char			 cbuf[14];	/* the actual command bytes */

	struct fas_dma_chain	 dma[MAXCHAIN];	/* DMA chain blocks */
	short			 max_link;	/* Maximum used of above */
	short			 cur_link;	/* Currently handled block */

	u_char			*buf;		/* Virtual adress of data */
	int			 len;		/* Bytes left to transfer */

	vm_offset_t		 dma_buf;	/* Current DMA adress */
	int			 dma_len;	/* Current DMA length */

	vm_offset_t		 dma_blk_ptr;	/* Current chain adress */
	int			 dma_blk_len;	/* Current chain length */
	u_char			 dma_blk_flg;	/* Current chain flags */

	u_char			 state;		/* Nexus state, see below */
	u_short			 flags;		/* Nexus flags, see below */

	short			 period;	/* Sync period to request */
	u_char			 offset;	/* Sync offset to request */

	u_char			 syncper;	/* FAS216 variable storage */
	u_char			 syncoff;	/* FAS216 variable storage */
	u_char			 config3;	/* FAS216 variable storage */

	u_char			 lun_unit;	/* (Lun<<4) | Unit of nexus */
	u_char			 status;	/* Status byte from unit*/

#ifdef FAS_NEED_VM_PATCH
	struct vm_link_data	 vm_link_data;	/* Optional vm FIX data */
#endif
};

/* SCSI nexus_states */
#define FAS_NS_IDLE		0	/* Nexus idle */
#define FAS_NS_SELECTED		1	/* Last command was a SELECT command */
#define FAS_NS_DATA_IN		2	/* Last command was a TRANSFER_INFO */
					/* command during a data in phase */
#define FAS_NS_DATA_OUT		3	/* Last command was a TRANSFER_INFO */
					/* command during a data out phase */
#define FAS_NS_STATUS		4	/* We have send a COMMAND_COMPLETE */
					/* command and are awaiting status */
#define FAS_NS_MSG_IN		5	/* Last phase was MESSAGE IN */
#define FAS_NS_MSG_OUT		6	/* Last phase was MESSAGE OUT */
#define FAS_NS_SVC		7	/* We have sent the command */
#define FAS_NS_DISCONNECTING	8	/* We have recieved a disconnect msg */
#define FAS_NS_DISCONNECTED	9	/* We are disconnected */
#define FAS_NS_RESELECTED	10	/* We was reselected */
#define FAS_NS_DONE		11	/* Done. Prephsase to FINISHED */
#define FAS_NS_FINISHED		12	/* Realy done. Call scsi_done */
#define FAS_NS_SENSE		13	/* We are requesting sense */
#define FAS_NS_RESET		14	/* We are reseting this unit */

/* SCSI nexus flags */
#define FAS_NF_UNIT_BUSY	0x0001	/* Unit is not available */

#define FAS_NF_SELECT_ME	0x0002	/* Nexus is set up, waiting for bus */

#define FAS_NF_REQUEST_SENSE	0x0004	/* We should request sense */
#define FAS_NF_SENSING		0x0008	/* We are sensing */

#define FAS_NF_HAS_MSG		0x0010	/* We have recieved a complete msg */

#define FAS_NF_DO_SDTR		0x0020	/* We should send a SDTR */
#define FAS_NF_SDTR_SENT	0x0040	/* We have sent a SDTR */
#define FAS_NF_SYNC_TESTED	0x0080	/* We have negotiated sync */

#define FAS_NF_RESET		0x0100	/* Reset this nexus */
#define FAS_NF_SYNC		0x0200	/* Not used??? */
#define FAS_NF_RESELECT		0x0400	/* Not used??? */
#define FAS_NF_IMMEDIATE	0x0800	/* We are operating from fasicmd */

#define FAS_NF_DEBUG		0x8000	/* As it says: DEBUG */

struct	fas_softc {
	struct	device		 sc_dev;	/* System struct */
	struct	scsi_link	 sc_link;	/* proto for sub devices */

	TAILQ_HEAD(,fas_pending) sc_xs_pending;
	TAILQ_HEAD(,fas_pending) sc_xs_free;
	struct	fas_pending 	 sc_xs_store[MAXPENDING];

	fas_regmap_p		 sc_fas;	/* FAS216 Address */
	void			*sc_spec;	/* Board-specific data */

#ifdef FAS_NEED_VM_PATCH
	u_char			*sc_vm_link;
#endif
	
	u_char			*sc_bump_va;	/* Bumpbuf virtual adr */
	vm_offset_t		 sc_bump_pa;	/* Bumpbuf physical adr */
	int			 sc_bump_sz;	/* Bumpbuf size */

/* Configuration registers, must be set BEFORE fasinitialize */
	u_char			 sc_clock_freq;
	u_short			 sc_timeout;
	u_char			 sc_host_id;
	u_char			 sc_config_flags;

/* Generic DMA functions */
	int		       (*sc_setup_dma)();
	int		       (*sc_build_dma_chain)();
	int		       (*sc_need_bump)();

/* Generic Led data */
	int			 sc_led_status;
	void		       (*sc_led)();

/* Nexus list */
	struct nexus		 sc_nexus[8];
	struct nexus		*sc_cur_nexus;

/* Current transfer data */
#ifdef FAS_NEED_VM_PATCH
	short			 sc_vm_link_pages;
#endif

	u_char			*sc_buf;	/* va */
	int			 sc_len;

	vm_offset_t		 sc_dma_buf;	/* pa */
	int			 sc_dma_len;
	vm_offset_t		 sc_dma_blk_ptr;
	int			 sc_dma_blk_len;
	short			 sc_dma_blk_flg;

	struct fas_dma_chain	*sc_chain;	/* Current DMA chain */
	short			 sc_max_link;
	short			 sc_cur_link;

/* Interrupt registers */
	u_char			 sc_status;
	u_char			 sc_interrupt;
	u_char			 sc_resel[2];

	u_char			 sc_units_disconnected;

/* Storage for FAS216 config registers (current values) */
	u_char			 sc_config1;
	u_char			 sc_config2;
	u_char			 sc_config3;
	u_char			 sc_clock_conv_fact;
	u_char			 sc_timeout_val;
	u_char			 sc_clock_period;

	u_char			 sc_msg_in[7];
	u_char			 sc_msg_in_len;

	u_char			 sc_msg_out[7];
	u_char			 sc_msg_out_len;

	u_char			 sc_unit;
	u_char			 sc_lun;
	u_char			 sc_flags;
};

#define FAS_DMA_READ		0
#define FAS_DMA_WRITE		1
#define FAS_DMA_CLEAR		2

/* sc_flags */
#define FAS_ACTIVE		0x01
#define FAS_DONT_WAIT		0x02
#ifdef FAS_NEED_VM_PATCH
#define FAS_HAS_VM_LINK		0x04
#endif

/* SCSI Selection modes */
#define FAS_SELECT	0x00	/* Normal selection: No sync, no resel */
#define FAS_SELECT_R	0x01	/* Reselection allowed */
#define FAS_SELECT_S	0x02	/* Synchronous transfer allowed */
#define FAS_SELECT_I	0x04	/* Selection for fasicmd */
#define FAS_SELECT_K	0x08	/* Send a BUS DEVICE RESET message (Kill) */

/* Nice abbreviations of the above */
#define FAS_SELECT_RS	(FAS_SELECT_R|FAS_SELECT_S)
#define FAS_SELECT_RI	(FAS_SELECT_R|FAS_SELECT_I)
#define FAS_SELECT_SI	(FAS_SELECT_S|FAS_SELECT_I)
#define FAS_SELECT_RSI	(FAS_SELECT_R|FAS_SELECT_S|FAS_SELECT_I)

/* sc_config_flags */
#define FAS_NO_SYNCH	0x01	/* Disable synchronous transfer */
#define FAS_NO_DMA	0x02	/* Do not use DMA! EVER! */
#define FAS_NO_RESELECT	0x04	/* Do not allow relesection */
#define FAS_SLOW_CABLE	0x08	/* Cable is "unsafe" for fast scsi-2 */
#define FAS_SLOW_START	0x10	/* There are slow starters on the bus */

void	fasinitialize __P((struct fas_softc *sc));
void	fas_minphys __P((struct buf *bp));
u_int	fas_adinfo __P((void));
int	fas_scsicmd __P((struct scsi_xfer *));

#endif /* _FASVAR_H_ */
